@@ -3,11 +3,12 @@ use solana_zk_token_sdk::encryption::{auth_encryption::AeCiphertext, elgamal::El
 pub use solana_zk_token_sdk::zk_token_proof_instruction::*;
 use {
     crate::{
-        check_program_account, extension::confidential_transfer::*, instruction::TokenInstruction,
+        check_program_account,
+        extension::confidential_transfer::*,
+        instruction::{encode_instruction, TokenInstruction},
     },
     bytemuck::{Pod, Zeroable},
-    num_derive::{FromPrimitive, ToPrimitive},
-    num_traits::{FromPrimitive, ToPrimitive},
+    num_enum::{IntoPrimitive, TryFromPrimitive},
     solana_program::{
         instruction::{AccountMeta, Instruction},
         program_error::ProgramError,
@@ -19,7 +20,7 @@ use {
 };
 
 /// Confidential Transfer extension instructions
-#[derive(Clone, Copy, Debug, FromPrimitive, ToPrimitive)]
+#[derive(Clone, Copy, Debug, TryFromPrimitive, IntoPrimitive)]
 #[repr(u8)]
 pub enum ConfidentialTransferInstruction {
     /// Initializes confidential transfers for a mint.
@@ -200,6 +201,28 @@ pub enum ConfidentialTransferInstruction {
     ///
     Transfer,
 
+    /// Transfer tokens confidentially with fee.
+    ///
+    ///   * Single owner/delegate
+    ///   1. `[writable]` The source SPL Token account.
+    ///   2. `[writable]` The destination SPL Token account.
+    ///   3. `[]` The token mint.
+    ///   4. `[]` Instructions sysvar.
+    ///   5. `[signer]` The single source account owner.
+    ///
+    ///   * Multisignature owner/delegate
+    ///   1. `[writable]` The source SPL Token account.
+    ///   2. `[writable]` The destination SPL Token account.
+    ///   3. `[]` The token mint.
+    ///   4. `[]` Instructions sysvar.
+    ///   5. `[]` The multisig  source account owner.
+    ///   6.. `[signer]` Required M signer accounts for the SPL Token Multisig account.
+    ///
+    /// Data expected by this instruction:
+    ///   `TransferWithFeeInstructionData`
+    ///
+    TransferWithFee,
+
     /// Applies the pending balance to the available balance, based on the history of `Deposit`
     /// and/or `Transfer` instructions.
     ///
@@ -353,7 +376,7 @@ pub enum ConfidentialTransferInstruction {
 #[repr(C)]
 pub struct ConfigureAccountInstructionData {
     /// The public key associated with the account
-    pub elgamal_pubkey: EncryptionPubkey,
+    pub encryption_pubkey: EncryptionPubkey,
     /// The decryptable balance (always 0) once the configure account succeeds
     pub decryptable_zero_balance: DecryptableBalance,
 }
@@ -403,6 +426,17 @@ pub struct TransferInstructionData {
     pub proof_instruction_offset: i8,
 }
 
+/// Data expected by `ConfidentialTransferInstruction::TransferWithFee`
+#[derive(Clone, Copy, Pod, Zeroable)]
+#[repr(C)]
+pub struct TransferWithFeeInstructionData {
+    /// The new source decryptable balance if the transfer succeeds
+    pub new_source_decryptable_available_balance: DecryptableBalance,
+    /// Relative location of the `ProofInstruction::VerifyTransfer` instruction to the
+    /// `Transfer` instruction in the transaction
+    pub proof_instruction_offset: i8,
+}
+
 /// Data expected by `ConfidentialTransferInstruction::ApplyPendingBalance`
 #[derive(Clone, Copy, Pod, Zeroable)]
 #[repr(C)]
@@ -434,40 +468,6 @@ pub struct WithdrawWithheldTokensFromAccountsData {
     pub proof_instruction_offset: i8,
 }
 
-pub(crate) fn decode_instruction_type(
-    input: &[u8],
-) -> Result<ConfidentialTransferInstruction, ProgramError> {
-    if input.is_empty() {
-        Err(ProgramError::InvalidInstructionData)
-    } else {
-        FromPrimitive::from_u8(input[0]).ok_or(ProgramError::InvalidInstructionData)
-    }
-}
-
-pub(crate) fn decode_instruction_data<T: Pod>(input: &[u8]) -> Result<&T, ProgramError> {
-    if input.is_empty() {
-        Err(ProgramError::InvalidInstructionData)
-    } else {
-        pod_from_bytes(&input[1..])
-    }
-}
-
-fn encode_instruction<T: Pod>(
-    token_program_id: &Pubkey,
-    accounts: Vec<AccountMeta>,
-    instruction_type: ConfidentialTransferInstruction,
-    instruction_data: &T,
-) -> Instruction {
-    let mut data = TokenInstruction::ConfidentialTransferExtension.pack();
-    data.push(ToPrimitive::to_u8(&instruction_type).unwrap());
-    data.extend_from_slice(bytemuck::bytes_of(instruction_data));
-    Instruction {
-        program_id: *token_program_id,
-        accounts,
-        data,
-    }
-}
-
 /// Create a `InitializeMint` instruction
 pub fn initialize_mint(
     token_program_id: &Pubkey,
@@ -479,6 +479,7 @@ pub fn initialize_mint(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::InitializeMint,
         ct_mint,
     ))
@@ -503,6 +504,7 @@ pub fn update_mint(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::UpdateMint,
         new_ct_mint,
     ))
@@ -514,7 +516,7 @@ pub fn configure_account(
     token_program_id: &Pubkey,
     token_account: &Pubkey,
     mint: &Pubkey,
-    elgamal_pubkey: ElGamalPubkey,
+    encryption_pubkey: ElGamalPubkey,
     decryptable_zero_balance: AeCiphertext,
     authority: &Pubkey,
     multisig_signers: &[&Pubkey],
@@ -533,9 +535,10 @@ pub fn configure_account(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::ConfigureAccount,
         &ConfigureAccountInstructionData {
-            elgamal_pubkey: elgamal_pubkey.into(),
+            encryption_pubkey: encryption_pubkey.into(),
             decryptable_zero_balance: decryptable_zero_balance.into(),
         },
     ))
@@ -557,6 +560,7 @@ pub fn approve_account(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::ApproveAccount,
         &(),
     ))
@@ -586,6 +590,7 @@ pub fn inner_empty_account(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::EmptyAccount,
         &EmptyAccountInstructionData {
             proof_instruction_offset,
@@ -640,6 +645,7 @@ pub fn deposit(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::Deposit,
         &DepositInstructionData {
             amount: amount.into(),
@@ -680,6 +686,7 @@ pub fn inner_withdraw(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::Withdraw,
         &WithdrawInstructionData {
             amount: amount.into(),
@@ -752,6 +759,7 @@ pub fn inner_transfer(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::Transfer,
         &TransferInstructionData {
             new_source_decryptable_available_balance,
@@ -788,6 +796,73 @@ pub fn transfer(
     ])
 }
 
+/// Create a inner `TransferWithFee` instruction
+///
+/// This instruction is suitable for use with a cross-program `invoke`
+#[allow(clippy::too_many_arguments)]
+pub fn inner_transfer_with_fee(
+    token_program_id: &Pubkey,
+    source_token_account: &Pubkey,
+    destination_token_account: &Pubkey,
+    mint: &Pubkey,
+    new_source_decryptable_available_balance: DecryptableBalance,
+    authority: &Pubkey,
+    multisig_signers: &[&Pubkey],
+    proof_instruction_offset: i8,
+) -> Result<Instruction, ProgramError> {
+    check_program_account(token_program_id)?;
+    let mut accounts = vec![
+        AccountMeta::new(*source_token_account, false),
+        AccountMeta::new(*destination_token_account, false),
+        AccountMeta::new_readonly(*mint, false),
+        AccountMeta::new_readonly(sysvar::instructions::id(), false),
+        AccountMeta::new_readonly(*authority, multisig_signers.is_empty()),
+    ];
+
+    for multisig_signer in multisig_signers.iter() {
+        accounts.push(AccountMeta::new_readonly(**multisig_signer, true));
+    }
+
+    Ok(encode_instruction(
+        token_program_id,
+        accounts,
+        TokenInstruction::ConfidentialTransferExtension,
+        ConfidentialTransferInstruction::TransferWithFee,
+        &TransferWithFeeInstructionData {
+            new_source_decryptable_available_balance,
+            proof_instruction_offset,
+        },
+    ))
+}
+
+/// Create a `Transfer` instruction
+#[allow(clippy::too_many_arguments)]
+#[cfg(not(target_arch = "bpf"))]
+pub fn transfer_with_fee(
+    token_program_id: &Pubkey,
+    source_token_account: &Pubkey,
+    destination_token_account: &Pubkey,
+    mint: &Pubkey,
+    new_source_decryptable_available_balance: AeCiphertext,
+    authority: &Pubkey,
+    multisig_signers: &[&Pubkey],
+    proof_data: &TransferWithFeeData,
+) -> Result<Vec<Instruction>, ProgramError> {
+    Ok(vec![
+        verify_transfer_with_fee(proof_data),
+        inner_transfer_with_fee(
+            token_program_id,
+            source_token_account,
+            destination_token_account,
+            mint,
+            new_source_decryptable_available_balance.into(),
+            authority,
+            multisig_signers,
+            -1,
+        )?, // calls check_program_account
+    ])
+}
+
 /// Create a inner `ApplyPendingBalance` instruction
 ///
 /// This instruction is suitable for use with a cross-program `invoke`
@@ -812,6 +887,7 @@ pub fn inner_apply_pending_balance(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::ApplyPendingBalance,
         &ApplyPendingBalanceData {
             expected_pending_balance_credit_counter: expected_pending_balance_credit_counter.into(),
@@ -860,6 +936,7 @@ fn enable_or_disable_balance_credits(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         instruction,
         &(),
     ))
@@ -923,6 +1000,7 @@ pub fn inner_withdraw_withheld_tokens_from_mint(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::WithdrawWithheldTokensFromMint,
         &WithdrawWithheldTokensFromMintData {
             proof_instruction_offset,
@@ -985,6 +1063,7 @@ pub fn inner_withdraw_withheld_tokens_from_accounts(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::WithdrawWithheldTokensFromAccounts,
         &WithdrawWithheldTokensFromAccountsData {
             proof_instruction_offset,
@@ -1033,6 +1112,7 @@ pub fn harvest_withheld_tokens_to_mint(
     Ok(encode_instruction(
         token_program_id,
         accounts,
+        TokenInstruction::ConfidentialTransferExtension,
         ConfidentialTransferInstruction::HarvestWithheldTokensToMint,
         &(),
     ))
