@@ -6,6 +6,7 @@ use {
             confidential_transfer::{instruction::*, *},
             StateWithExtensions, StateWithExtensionsMut,
         },
+        instruction::{decode_instruction_data, decode_instruction_type},
         processor::Processor,
         state::{Account, Mint},
     },
@@ -32,6 +33,10 @@ fn decode_proof_instruction<T: Pod>(
     expected: ProofInstruction,
     instruction: &Instruction,
 ) -> Result<&T, ProgramError> {
+    if ProofInstruction::decode_type(&instruction.data) != Some(expected) {
+        msg!("decode type failed ----------------------------------");
+    }
+
     if instruction.program_id != zk_token_proof_program::id()
         || ProofInstruction::decode_type(&instruction.data) != Some(expected)
     {
@@ -90,7 +95,7 @@ fn process_configure_account(
     program_id: &Pubkey,
     accounts: &[AccountInfo],
     ConfigureAccountInstructionData {
-        elgamal_pubkey,
+        encryption_pubkey,
         decryptable_zero_balance,
     }: &ConfigureAccountInstructionData,
 ) -> ProgramResult {
@@ -126,7 +131,7 @@ fn process_configure_account(
     let mut confidential_transfer_account =
         token_account.init_extension::<ConfidentialTransferAccount>()?;
     confidential_transfer_account.approved = confidential_transfer_mint.auto_approve_new_accounts;
-    confidential_transfer_account.pubkey_elgamal = *elgamal_pubkey;
+    confidential_transfer_account.encryption_pubkey = *encryption_pubkey;
 
     /*
         An ElGamal ciphertext is of the form
@@ -240,11 +245,6 @@ fn process_empty_account(
     }
 
     confidential_transfer_account.available_balance = EncryptedBalance::zeroed();
-
-    if confidential_transfer_account.withheld_amount != EncryptedWithheldAmount::zeroed() {
-        msg!("Withheld amount is not zero");
-        return Err(ProgramError::InvalidAccountData);
-    }
     confidential_transfer_account.closable()?;
 
     Ok(())
@@ -481,12 +481,12 @@ fn process_transfer(
     if let Ok(transfer_fee_config) = mint.get_extension::<TransferFeeConfig>() {
         // mint is extended for fees
         let proof_data = decode_proof_instruction::<TransferWithFeeData>(
-            ProofInstruction::VerifyTransfer,
+            ProofInstruction::VerifyTransferWithFee,
             &previous_instruction,
         )?;
 
-        if proof_data.transfer_with_fee_pubkeys.pubkey_auditor
-            != confidential_transfer_mint.pubkey_auditor
+        if proof_data.transfer_with_fee_pubkeys.auditor_pubkey
+            != confidential_transfer_mint.auditor_pubkey
         {
             return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
         }
@@ -494,8 +494,8 @@ fn process_transfer(
         // `withdraw_withheld_authority` ElGamal pubkey in proof data and mint must match
         if proof_data
             .transfer_with_fee_pubkeys
-            .pubkey_withdraw_withheld_authority
-            != confidential_transfer_mint.pubkey_withdraw_withheld_authority
+            .withdraw_withheld_authority_pubkey
+            != confidential_transfer_mint.withdraw_withheld_authority_pubkey
         {
             return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
         }
@@ -532,11 +532,11 @@ fn process_transfer(
 
         let ciphertext_lo = EncryptedBalance::from((
             proof_data.ciphertext_lo.commitment,
-            proof_data.ciphertext_lo.handle_source,
+            proof_data.ciphertext_lo.source_handle,
         ));
         let ciphertext_hi = EncryptedBalance::from((
             proof_data.ciphertext_hi.commitment,
-            proof_data.ciphertext_hi.handle_source,
+            proof_data.ciphertext_hi.source_handle,
         ));
 
         process_source_for_transfer(
@@ -545,7 +545,7 @@ fn process_transfer(
             mint_info,
             authority_info,
             account_info_iter.as_slice(),
-            &proof_data.transfer_with_fee_pubkeys.pubkey_source,
+            &proof_data.transfer_with_fee_pubkeys.source_pubkey,
             &ciphertext_lo,
             &ciphertext_hi,
             new_source_decryptable_available_balance,
@@ -553,10 +553,10 @@ fn process_transfer(
         process_destination_for_transfer(
             destination_token_account_info,
             mint_info,
-            &proof_data.transfer_with_fee_pubkeys.pubkey_dest,
+            &proof_data.transfer_with_fee_pubkeys.destination_pubkey,
             &ciphertext_lo,
             &ciphertext_hi,
-            Some(proof_data.ciphertext_fee),
+            Some(proof_data.fee_ciphertext),
         )?;
     } else {
         // mint is not extended for fees
@@ -565,25 +565,26 @@ fn process_transfer(
             &previous_instruction,
         )?;
 
-        if proof_data.transfer_pubkeys.pubkey_auditor != confidential_transfer_mint.pubkey_auditor {
+        if proof_data.transfer_pubkeys.auditor_pubkey != confidential_transfer_mint.auditor_pubkey {
             return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
         }
 
         let ciphertext_lo = EncryptedBalance::from((
             proof_data.ciphertext_lo.commitment,
-            proof_data.ciphertext_lo.handle_source,
+            proof_data.ciphertext_lo.source_handle,
         ));
         let ciphertext_hi = EncryptedBalance::from((
             proof_data.ciphertext_hi.commitment,
-            proof_data.ciphertext_hi.handle_source,
+            proof_data.ciphertext_hi.source_handle,
         ));
+
         process_source_for_transfer(
             program_id,
             token_account_info,
             mint_info,
             authority_info,
             account_info_iter.as_slice(),
-            &proof_data.transfer_pubkeys.pubkey_source,
+            &proof_data.transfer_pubkeys.source_pubkey,
             &ciphertext_lo,
             &ciphertext_hi,
             new_source_decryptable_available_balance,
@@ -592,7 +593,7 @@ fn process_transfer(
         process_destination_for_transfer(
             destination_token_account_info,
             mint_info,
-            &proof_data.transfer_pubkeys.pubkey_dest,
+            &proof_data.transfer_pubkeys.destination_pubkey,
             &ciphertext_lo,
             &ciphertext_hi,
             None,
@@ -610,9 +611,9 @@ fn process_source_for_transfer(
     mint_info: &AccountInfo,
     authority_info: &AccountInfo,
     signers: &[AccountInfo],
-    elgamal_pubkey_source: &EncryptionPubkey,
-    ciphertext_lo_source: &EncryptedBalance,
-    ciphertext_hi_source: &EncryptedBalance,
+    source_encryption_pubkey: &EncryptionPubkey,
+    source_ciphertext_lo: &EncryptedBalance,
+    source_ciphertext_hi: &EncryptedBalance,
     new_source_decryptable_available_balance: DecryptableBalance,
 ) -> ProgramResult {
     check_program_account(token_account_info.owner)?;
@@ -639,15 +640,15 @@ fn process_source_for_transfer(
     let mut confidential_transfer_account =
         token_account.get_extension_mut::<ConfidentialTransferAccount>()?;
     confidential_transfer_account.approved()?;
-    if *elgamal_pubkey_source != confidential_transfer_account.pubkey_elgamal {
+    if *source_encryption_pubkey != confidential_transfer_account.encryption_pubkey {
         return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
     }
 
     let new_source_available_balance = {
         ops::subtract_with_lo_hi(
             &confidential_transfer_account.available_balance,
-            ciphertext_lo_source,
-            ciphertext_hi_source,
+            source_ciphertext_lo,
+            source_ciphertext_hi,
         )
         .ok_or(ProgramError::InvalidInstructionData)?
     };
@@ -663,9 +664,9 @@ fn process_source_for_transfer(
 fn process_destination_for_transfer(
     destination_token_account_info: &AccountInfo,
     mint_info: &AccountInfo,
-    elgamal_pubkey_dest: &EncryptionPubkey,
-    ciphertext_lo_destination: &EncryptedBalance,
-    ciphertext_hi_destination: &EncryptedBalance,
+    destination_encryption_pubkey: &EncryptionPubkey,
+    destination_ciphertext_lo: &EncryptedBalance,
+    destination_ciphertext_hi: &EncryptedBalance,
     encrypted_fee: Option<EncryptedFee>,
 ) -> ProgramResult {
     check_program_account(destination_token_account_info.owner)?;
@@ -689,14 +690,15 @@ fn process_destination_for_transfer(
         return Err(TokenError::ConfidentialTransferDepositsAndTransfersDisabled.into());
     }
 
-    if *elgamal_pubkey_dest != destination_confidential_transfer_account.pubkey_elgamal {
+    if *destination_encryption_pubkey != destination_confidential_transfer_account.encryption_pubkey
+    {
         return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
     }
 
     let new_destination_pending_balance = ops::add_with_lo_hi(
         &destination_confidential_transfer_account.pending_balance,
-        ciphertext_lo_destination,
-        ciphertext_hi_destination,
+        destination_ciphertext_lo,
+        destination_ciphertext_hi,
     )
     .ok_or(ProgramError::InvalidInstructionData)?;
 
@@ -711,10 +713,10 @@ fn process_destination_for_transfer(
     // update destination account withheld fees
     if let Some(ciphertext_fee) = encrypted_fee {
         let ciphertext_fee_destination: EncryptedWithheldAmount =
-            (ciphertext_fee.commitment, ciphertext_fee.handle_dest).into();
+            (ciphertext_fee.commitment, ciphertext_fee.destination_handle).into();
         let ciphertext_fee_withheld_authority: EncryptedWithheldAmount = (
             ciphertext_fee.commitment,
-            ciphertext_fee.handle_withdraw_withheld_authority,
+            ciphertext_fee.withdraw_withheld_authority_handle,
         )
             .into();
 
@@ -872,19 +874,20 @@ fn process_withdraw_withheld_tokens_from_mint(
     )?;
 
     // withdraw withheld authority ElGamal pubkey should match in the proof data and mint
-    if proof_data.pubkey_withdraw_withheld_authority
-        != confidential_transfer_mint.pubkey_withdraw_withheld_authority
+    if proof_data.withdraw_withheld_authority_pubkey
+        != confidential_transfer_mint.withdraw_withheld_authority_pubkey
     {
         return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
     }
 
     // destination ElGamal pubkey should match in the proof data and destination account
-    if proof_data.pubkey_dest != destination_confidential_transfer_account.pubkey_elgamal {
+    if proof_data.destination_pubkey != destination_confidential_transfer_account.encryption_pubkey
+    {
         return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
     }
 
     // withheld amount ciphertext must match in the proof data and mint
-    if proof_data.ciphertext_withdraw_withheld_authority
+    if proof_data.withdraw_withheld_authority_ciphertext
         != confidential_transfer_mint.withheld_amount
     {
         return Err(TokenError::ConfidentialTransferBalanceMismatch.into());
@@ -894,7 +897,7 @@ fn process_withdraw_withheld_tokens_from_mint(
     // This amount should be added to the destination pending balance.
     let new_destination_pending_balance = ops::add(
         &destination_confidential_transfer_account.pending_balance,
-        &proof_data.ciphertext_dest,
+        &proof_data.destination_ciphertext,
     )
     .ok_or(ProgramError::InvalidInstructionData)?;
 
@@ -994,19 +997,20 @@ fn process_withdraw_withheld_tokens_from_accounts(
 
     // withdraw withheld authority ElGamal pubkey should match in the proof data and mint
     let confidential_transfer_mint = mint.get_extension_mut::<ConfidentialTransferMint>()?;
-    if proof_data.pubkey_withdraw_withheld_authority
-        != confidential_transfer_mint.pubkey_withdraw_withheld_authority
+    if proof_data.withdraw_withheld_authority_pubkey
+        != confidential_transfer_mint.withdraw_withheld_authority_pubkey
     {
         return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
     }
 
     // destination ElGamal pubkey should match in the proof data and destination account
-    if proof_data.pubkey_dest != destination_confidential_transfer_account.pubkey_elgamal {
+    if proof_data.destination_pubkey != destination_confidential_transfer_account.encryption_pubkey
+    {
         return Err(TokenError::ConfidentialTransferElGamalPubkeyMismatch.into());
     }
 
     // withheld amount ciphertext must match in the proof data and mint
-    if proof_data.ciphertext_withdraw_withheld_authority != aggregate_withheld_amount {
+    if proof_data.withdraw_withheld_authority_ciphertext != aggregate_withheld_amount {
         return Err(TokenError::ConfidentialTransferBalanceMismatch.into());
     }
 
@@ -1158,6 +1162,23 @@ pub(crate) fn process_instruction(
             #[cfg(not(feature = "zk-ops"))]
             Err(ProgramError::InvalidInstructionData)
         }
+        ConfidentialTransferInstruction::TransferWithFee => {
+            msg!("ConfidentialTransferInstruction::TransferWithFee");
+            #[cfg(feature = "zk-ops")]
+            {
+                let data = decode_instruction_data::<TransferWithFeeInstructionData>(input)?;
+                process_transfer(
+                    program_id,
+                    accounts,
+                    data.new_source_decryptable_available_balance,
+                    data.proof_instruction_offset as i64,
+                )
+            }
+            #[cfg(not(feature = "zk-ops"))]
+            {
+                Err(ProgramError::InvalidInstructionData)
+            }
+        }
         ConfidentialTransferInstruction::ApplyPendingBalance => {
             msg!("ConfidentialTransferInstruction::ApplyPendingBalance");
             #[cfg(feature = "zk-ops")]
