@@ -5,7 +5,9 @@ use {
         check_program_account, check_spl_token_program_account,
         error::TokenError,
         extension::{transfer_fee::instruction::TransferFeeInstruction, ExtensionType},
+        pod::{pod_from_bytes, pod_get_packed_len},
     },
+    bytemuck::Pod,
     solana_program::{
         instruction::{AccountMeta, Instruction},
         program_error::ProgramError,
@@ -13,7 +15,10 @@ use {
         pubkey::{Pubkey, PUBKEY_BYTES},
         system_program, sysvar,
     },
-    std::{convert::TryInto, mem::size_of},
+    std::{
+        convert::{TryFrom, TryInto},
+        mem::size_of,
+    },
 };
 
 /// Minimum number of multisignature signers (min N)
@@ -910,12 +915,14 @@ pub enum AuthorityType {
     FreezeAccount,
     /// Owner of a given token account
     AccountOwner,
-    /// Authority to close a mint or token account
+    /// Authority to close a token account
     CloseAccount,
     /// Authority to set the transfer fee
     TransferFeeConfig,
     /// Authority to withdraw withheld tokens from a mint
     WithheldWithdraw,
+    /// Authority to close a mint account
+    CloseMint,
 }
 
 impl AuthorityType {
@@ -927,6 +934,7 @@ impl AuthorityType {
             AuthorityType::CloseAccount => 3,
             AuthorityType::TransferFeeConfig => 4,
             AuthorityType::WithheldWithdraw => 5,
+            AuthorityType::CloseMint => 6,
         }
     }
 
@@ -938,6 +946,7 @@ impl AuthorityType {
             3 => Ok(AuthorityType::CloseAccount),
             4 => Ok(AuthorityType::TransferFeeConfig),
             5 => Ok(AuthorityType::WithheldWithdraw),
+            6 => Ok(AuthorityType::CloseMint),
             _ => Err(TokenError::InvalidInstruction.into()),
         }
     }
@@ -1680,6 +1689,42 @@ pub fn is_valid_signer_index(index: usize) -> bool {
     (MIN_SIGNERS..=MAX_SIGNERS).contains(&index)
 }
 
+/// Utility function for decoding just the instruction type
+pub fn decode_instruction_type<T: TryFrom<u8>>(input: &[u8]) -> Result<T, ProgramError> {
+    if input.is_empty() {
+        Err(ProgramError::InvalidInstructionData)
+    } else {
+        T::try_from(input[0]).map_err(|_| TokenError::InvalidInstruction.into())
+    }
+}
+
+/// Utility function for decoding instruction data
+pub fn decode_instruction_data<T: Pod>(input: &[u8]) -> Result<&T, ProgramError> {
+    if input.len() != pod_get_packed_len::<T>().saturating_add(1) {
+        Err(ProgramError::InvalidInstructionData)
+    } else {
+        pod_from_bytes(&input[1..])
+    }
+}
+
+/// Utility function for encoding instruction data
+pub(crate) fn encode_instruction<T: Into<u8>, D: Pod>(
+    token_program_id: &Pubkey,
+    accounts: Vec<AccountMeta>,
+    token_instruction_type: TokenInstruction,
+    instruction_type: T,
+    instruction_data: &D,
+) -> Instruction {
+    let mut data = token_instruction_type.pack();
+    data.push(T::into(instruction_type));
+    data.extend_from_slice(bytemuck::bytes_of(instruction_data));
+    Instruction {
+        program_id: *token_program_id,
+        accounts,
+        data,
+    }
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
@@ -1949,4 +1994,234 @@ mod test {
         let unpacked = TokenInstruction::unpack(&expect).unwrap();
         assert_eq!(unpacked, check);
     }
+
+    macro_rules! test_instruction {
+        ($a:ident($($b:tt)*)) => {
+            let instruction_v3 = spl_token::instruction::$a($($b)*).unwrap();
+            let instruction_2022 = $a($($b)*).unwrap();
+            assert_eq!(instruction_v3, instruction_2022);
+        }
+    }
+
+    #[test]
+    fn test_v3_compatibility() {
+        let token_program_id = spl_token::id();
+        let mint_pubkey = Pubkey::new_unique();
+        let mint_authority_pubkey = Pubkey::new_unique();
+        let freeze_authority_pubkey = Pubkey::new_unique();
+        let decimals = 9u8;
+
+        let account_pubkey = Pubkey::new_unique();
+        let owner_pubkey = Pubkey::new_unique();
+
+        let multisig_pubkey = Pubkey::new_unique();
+        let signer_pubkeys_vec = vec![Pubkey::new_unique(); MAX_SIGNERS];
+        let signer_pubkeys = signer_pubkeys_vec.iter().collect::<Vec<_>>();
+        let m = 10u8;
+
+        let source_pubkey = Pubkey::new_unique();
+        let destination_pubkey = Pubkey::new_unique();
+        let authority_pubkey = Pubkey::new_unique();
+        let amount = 1_000_000_000_000;
+
+        let delegate_pubkey = Pubkey::new_unique();
+        let owned_pubkey = Pubkey::new_unique();
+        let new_authority_pubkey = Pubkey::new_unique();
+
+        let ui_amount = "100000.00";
+
+        test_instruction!(initialize_mint(
+            &token_program_id,
+            &mint_pubkey,
+            &mint_authority_pubkey,
+            None,
+            decimals,
+        ));
+        test_instruction!(initialize_mint2(
+            &token_program_id,
+            &mint_pubkey,
+            &mint_authority_pubkey,
+            Some(&freeze_authority_pubkey),
+            decimals,
+        ));
+
+        test_instruction!(initialize_account(
+            &token_program_id,
+            &account_pubkey,
+            &mint_pubkey,
+            &owner_pubkey,
+        ));
+        test_instruction!(initialize_account2(
+            &token_program_id,
+            &account_pubkey,
+            &mint_pubkey,
+            &owner_pubkey,
+        ));
+        test_instruction!(initialize_account3(
+            &token_program_id,
+            &account_pubkey,
+            &mint_pubkey,
+            &owner_pubkey,
+        ));
+        test_instruction!(initialize_multisig(
+            &token_program_id,
+            &multisig_pubkey,
+            &signer_pubkeys,
+            m,
+        ));
+        test_instruction!(initialize_multisig2(
+            &token_program_id,
+            &multisig_pubkey,
+            &signer_pubkeys,
+            m,
+        ));
+        #[allow(deprecated)]
+        {
+            test_instruction!(transfer(
+                &token_program_id,
+                &source_pubkey,
+                &destination_pubkey,
+                &authority_pubkey,
+                &signer_pubkeys,
+                amount
+            ));
+        }
+        test_instruction!(transfer_checked(
+            &token_program_id,
+            &source_pubkey,
+            &mint_pubkey,
+            &destination_pubkey,
+            &authority_pubkey,
+            &signer_pubkeys,
+            amount,
+            decimals,
+        ));
+        test_instruction!(approve(
+            &token_program_id,
+            &source_pubkey,
+            &delegate_pubkey,
+            &owner_pubkey,
+            &signer_pubkeys,
+            amount
+        ));
+        test_instruction!(approve_checked(
+            &token_program_id,
+            &source_pubkey,
+            &mint_pubkey,
+            &delegate_pubkey,
+            &owner_pubkey,
+            &signer_pubkeys,
+            amount,
+            decimals
+        ));
+        test_instruction!(revoke(
+            &token_program_id,
+            &source_pubkey,
+            &owner_pubkey,
+            &signer_pubkeys,
+        ));
+
+        // set_authority
+        {
+            let instruction_v3 = spl_token::instruction::set_authority(
+                &token_program_id,
+                &owned_pubkey,
+                Some(&new_authority_pubkey),
+                spl_token::instruction::AuthorityType::AccountOwner,
+                &owner_pubkey,
+                &signer_pubkeys,
+            )
+            .unwrap();
+            let instruction_2022 = set_authority(
+                &token_program_id,
+                &owned_pubkey,
+                Some(&new_authority_pubkey),
+                AuthorityType::AccountOwner,
+                &owner_pubkey,
+                &signer_pubkeys,
+            )
+            .unwrap();
+            assert_eq!(instruction_v3, instruction_2022);
+        }
+
+        test_instruction!(mint_to(
+            &token_program_id,
+            &mint_pubkey,
+            &account_pubkey,
+            &owner_pubkey,
+            &signer_pubkeys,
+            amount,
+        ));
+        test_instruction!(mint_to_checked(
+            &token_program_id,
+            &mint_pubkey,
+            &account_pubkey,
+            &owner_pubkey,
+            &signer_pubkeys,
+            amount,
+            decimals,
+        ));
+        test_instruction!(burn(
+            &token_program_id,
+            &account_pubkey,
+            &mint_pubkey,
+            &authority_pubkey,
+            &signer_pubkeys,
+            amount,
+        ));
+        test_instruction!(burn_checked(
+            &token_program_id,
+            &account_pubkey,
+            &mint_pubkey,
+            &authority_pubkey,
+            &signer_pubkeys,
+            amount,
+            decimals,
+        ));
+        test_instruction!(close_account(
+            &token_program_id,
+            &account_pubkey,
+            &destination_pubkey,
+            &owner_pubkey,
+            &signer_pubkeys,
+        ));
+        test_instruction!(freeze_account(
+            &token_program_id,
+            &account_pubkey,
+            &mint_pubkey,
+            &owner_pubkey,
+            &signer_pubkeys,
+        ));
+        test_instruction!(thaw_account(
+            &token_program_id,
+            &account_pubkey,
+            &mint_pubkey,
+            &owner_pubkey,
+            &signer_pubkeys,
+        ));
+        test_instruction!(sync_native(&token_program_id, &account_pubkey,));
+
+        // get_account_data_size
+        {
+            let instruction_v3 =
+                spl_token::instruction::get_account_data_size(&token_program_id, &mint_pubkey)
+                    .unwrap();
+            let instruction_2022 =
+                get_account_data_size(&token_program_id, &mint_pubkey, &[]).unwrap();
+            assert_eq!(instruction_v3, instruction_2022);
+        }
+
+        test_instruction!(initialize_immutable_owner(
+            &token_program_id,
+            &account_pubkey,
+        ));
+
+        test_instruction!(amount_to_ui_amount(&token_program_id, &mint_pubkey, amount,));
+
+        test_instruction!(ui_amount_to_amount(
+            &token_program_id,
+            &mint_pubkey,
+            ui_amount,
+        ));
+    }
 }
